from enum import IntEnum
import logging
import struct
from typing import List, Dict, Tuple, Any, Optional, cast

import usb.core  # type: ignore
import usb.util  # type: ignore
import usb.legacy  # type: ignore


log = logging.getLogger('picoslave')


def bytestr(b: bytes) -> str:
    return ' '.join(f'{x:02X}' for x in b)


class DeviceNotFound(Exception):
    """error when a specified device was not found"""


class UsbError(Exception):
    """error when communication with the device was not possible"""


class ProtocolError(Exception):
    """when unexpected respnses are received from the device"""


class UsbPicoSlave:

    ID_VENDOR = 0xC001
    ID_PRODUCT = 0x012C
    OUT_EP_ADDR = 0x04
    IN_EP_ADDR = 0x85

    def __init__(self, device: usb.core.Device) -> None:
        assert device is not None

        cfg = device.get_active_configuration()
        assert cfg is not None, 'no active configuration on device'

        intf = usb.util.find_descriptor(cfg, bInterfaceClass=usb.legacy.CLASS_VENDOR_SPEC)
        assert intf is not None, "couldn't find PicoSlave interface on device"

        outep = usb.util.find_descriptor(intf, bEndpointAddress=self.OUT_EP_ADDR)
        assert outep is not None, f"OUT endpoint (0x{self.OUT_EP_ADDR:02X}) not found on device"

        inep = usb.util.find_descriptor(intf, bEndpointAddress=self.IN_EP_ADDR)
        assert outep is not None, f"IN endpoint (0x{self.IN_EP_ADDR:02X}) not found on device"

        self._device = device
        self._inep = inep
        self._outep = outep

    def close(self) -> None:
        self._device.reset()

    def write(self, data: bytes) -> None:
        """raises: UsbError"""
        try:
            wire_hdr = struct.pack('I', len(data))
            self._outep.write(wire_hdr + data)
        except usb.core.USBTimeoutError as e:
            raise UsbError(f'No response from device ({e})')

    def read(self) -> bytes:
        """raises: UsbError"""
        try:
            packet = self._inep.read(8192)
            len_hdr = struct.unpack('I', packet[:4])[0]
            data = packet[4:]
            log.debug(f'[rx]: len_hdr: {len_hdr}, len(data): {len(data)}')
            if len_hdr != len(data):
                raise ProtocolError('received wrong number of bytes')
            return bytes(list(data))
        except usb.core.USBTimeoutError as e:
            raise UsbError(f'No response from device ({e})')

    @classmethod
    def _scan(cls) -> List[usb.core.Device]:
        devs = usb.core.find(idVendor=cls.ID_VENDOR, idProduct=cls.ID_PRODUCT, find_all=True)
        devices: List[usb.core.Device] = []
        for d in devs:
            try:
                # try to access the serial number.
                # this gives an error when the device is in a weird state
                log.debug(f'found device with serial number: {d.serial_number}')
            except Exception as e:
                log.warning(f"couldn't read from device: {e}")
                continue
            devices.append(d)
        return sorted(devices, key=lambda d: cast(int, d.serial_number))

    @classmethod
    def scan(cls) -> List[usb.core.Device]:
        devices = cls._scan()
        vid_pid = f'{cls.ID_VENDOR:04x}:{cls.ID_PRODUCT:04x}'
        id = 1
        for d in devices:
            usb_path = f'{d.bus}.{".".join(str(p) for p in d.port_numbers)}:{d.address}'
            print(f'{id:>2}  {d.serial_number:<16}  {usb_path:>12}-{vid_pid:<9}'
                  f'  {d.manufacturer} {d.product} ')
            id += 1
        if not devices:
            print(f'no PicoSlave device found with {vid_pid}')
        return devices

    @classmethod
    def find(cls, index: Optional[int] = None, serial: Optional[str] = None) -> 'UsbPicoSlave':
        assert index is None or serial is None, "'index' and 'serial' must not be both set"

        selected_device: usb.core.Device = None
        devices = cls._scan()

        try:
            if index is not None:
                assert index < len(devices), f'No device with index "{index}"'
                selected_device = devices[index]
            elif serial is not None:
                matching = [dev for dev in devices if dev.serial_number == serial]
                assert matching, f'No device found with serial number "{serial}"'
                assert len(matching) == 1, \
                    f'Multiple devices found with serial number "{serial}". '\
                    'Select with "index" filter'
                selected_device = matching[0]
            else:
                assert len(devices) > 0, 'No device found'
                assert len(devices) == 1, \
                    'Multiple devices found. Select with "index" or "serial" filter.'
                selected_device = devices[0]
            assert selected_device is not None, 'FATAL: No device found (unknown reason)'
        except AssertionError as e:
            raise DeviceNotFound(e) from e
        return UsbPicoSlave(selected_device)


class PicoSlave:
    class Protocol:

        MAX_RW_SIZE = 32

        class Command(IntEnum):
            CONFIG = 0xA0
            READ = 0xA1
            WRITE = 0xA2
            CLEAR = 0xA3
            STAT = 0xA4
            INFO = 0xB0
            RESET = 0xBF

        class ResponseCode(IntEnum):
            OK = 0
            CRC_ERROR = 1
            INVALID_PACKET = 2
            INVALID_REQUEST = 3
            INVALID_INTERFACE = 4
            INVALID_ADDRESS = 5
            INVALID_SIZE = 6
            INVALID_WIDTH = 7
            MEMORY_ERROR = 8
            OPERATION_FAILED = 9

        @staticmethod
        def crc16(data: bytes, poly: int = 0x8408) -> int:
            """CRC-16-CCITT Algorithm
            source: https://gist.github.com/oysstu/68072c44c02879a2abf94ef350d1c7c6"""
            crc = 0xFFFF
            for b in data:
                cur_byte = 0xFF & b
                for _ in range(0, 8):
                    if (crc & 0x0001) ^ (cur_byte & 0x0001):
                        crc = (crc >> 1) ^ poly
                    else:
                        crc >>= 1
                    cur_byte >>= 1
            crc = (~crc & 0xFFFF)
            crc = (crc << 8) | ((crc >> 8) & 0xFF)
            return crc & 0xFFFF

        @classmethod
        def packet(cls, cmd: Command, iface: int, addr: int, size: int = 0,
                   data: Optional[bytes] = None, mem_width: int = 0) -> bytes:
            assert size is None or size < 2**16
            assert addr < 2**16, 'addr must be in range [0..2**16]'
            if cmd == cls.Command.CONFIG:
                assert data is None, 'CONFIG must not have data'
                assert size > 0, 'CONFIG needs size > 0'
                data = struct.pack('B', mem_width)
            elif cmd == cls.Command.READ or cmd == cls.Command.STAT:
                assert size > 0, 'READ/STAT packet needs size > 0'
                assert data is None, 'READ/STAT must not have data'
            elif cmd == cls.Command.WRITE:
                assert data is not None, 'WRITE packet needs data'
                size = len(data)
            elif cmd == cls.Command.CLEAR:
                assert size == 0, 'CLEAR size must be 0'
                assert data is None, 'CLEAR must not have data'
            elif cmd == cls.Command.RESET:
                assert size == 0, 'RESET size must be 0'
                assert data is None, 'RESET must not have data'
            elif cmd == cls.Command.INFO:
                assert size == 0, 'INFO size must be 0'
                assert data is None, 'INFO must not have data'
            else:
                assert False, f'Unsupported CMD: {cmd}'
            header = bytes([int(cmd), iface]) + struct.pack('H', addr) + struct.pack('H', size)
            packet = header + (data or bytes([]))
            crc = struct.pack('H', cls.crc16(packet))
            return packet + crc

        @classmethod
        def verify(cls, packet: bytes) -> bytes:
            # check that the packet is long enough to make any sense
            assert len(packet) >= 5, f'invalid received packet size: {len(packet)}'

            # check CRC before looking into the packet
            crc_calc = cls.crc16(packet[:-2])
            crc_rx = struct.unpack('H', packet[-2:])[0]
            assert crc_calc == crc_rx, 'CRC mismatch'
            log.debug('rx packet CRC check successful')
            packet_data = packet[:-2]

            # check the result code
            try:
                code = packet_data[0]
                response_code = cls.ResponseCode(code)
                assert response_code == cls.ResponseCode.OK, \
                    f'received error code {code} ({response_code.name})'
                # FIXME: this should be a testable exception
            except ValueError:
                assert False, f'received unknown response code: {code}'

            # check the packet data size (-3 is the header size)
            size = struct.unpack('H', packet_data[1:3])[0]
            assert size == len(packet_data) - 3, \
                f'received packet size mismatch: size={size}, data={len(packet_data) - 3}'
            return packet_data[3:]

    def __init__(self, index: Optional[int] = None, serial: Optional[str] = None) -> None:
        assert index is None or serial is None, "'index' and 'serial' must not be both set"
        self._usb = UsbPicoSlave.find(index=index, serial=serial)

    def close(self) -> None:
        self._usb.close()

    def _txrx(self, packet: bytes) -> bytes:
        """raises: UsbError"""

        log.debug(f'tx-packet: {bytestr(packet)}')
        try:
            self._usb.write(packet)
        except usb.core.USBError as e:
            raise UsbError(e)

        try:
            response = self._usb.read()
        except usb.core.USBError as e:
            raise UsbError(e)
        log.debug(f'rx-packet: {bytestr(response)}')

        try:
            rx_data = self.Protocol.verify(response)
        except AssertionError as e:
            raise ProtocolError(e)
        log.debug(f'rx-data: {bytestr(rx_data)}')
        return rx_data

    def config(self, iface: int, slave_address: int, mem_size: int = 256,
               mem_width: int = 1) -> None:
        """Initialize or re-initialize the slave for a given I2C address.
        Address 0 deactivates the slave. """
        assert mem_size > 0  # FIXME [#12]: can't test currently... and mem_size <= 256
        assert mem_width in [1, 2, 4]
        log.info(f'config: iface={iface}, slave_address={slave_address:02X}h, '
                 f'mem_size={mem_size}, mem_width={mem_width}')
        packet = self.Protocol.packet(self.Protocol.Command.CONFIG,
                                      iface, slave_address, mem_size, mem_width=mem_width)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'CONFIG received an unexpected response: {bytestr(rsp)}')

    def read(self, iface: int, mem_addr: int, size: int) -> bytes:
        log.info(f'read: iface={iface}, mem_addr={mem_addr:02X}h, size={size}')
        packet = self.Protocol.packet(self.Protocol.Command.READ, iface, mem_addr, size)
        rsp = self._txrx(packet)
        assert len(rsp) > 0, 'received no data'
        return rsp

    def write(self, iface: int, mem_addr: int, data: bytes) -> None:
        log.info(f'write: iface={iface}, mem_addr={mem_addr:02X}h, data_size={len(data)}, '
                 f'data="{bytestr(data)}"')
        packet = self.Protocol.packet(self.Protocol.Command.WRITE, iface, mem_addr, len(data), data)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'WRITE received an unexpected response: {bytestr(rsp)}')

    def clear(self, iface: int, reset_value: int = 0) -> None:
        log.info(f'clear: iface={iface}, reset_value={reset_value}')
        packet = self.Protocol.packet(self.Protocol.Command.CLEAR, iface, reset_value)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'CLEAR received an unexpected response: {bytestr(rsp)}')

    def statistics(self, iface: int, mem_addr: int, size: int) -> Dict[int, Tuple[int, int]]:
        log.info(f'stat: iface={iface}, mem_addr={mem_addr:02X}h, size={size}')
        packet = self.Protocol.packet(self.Protocol.Command.STAT, iface, mem_addr, size)
        rsp = self._txrx(packet)
        assert len(rsp) > 0, 'received no data'
        assert len(rsp) % 8 == 0, f'received data size is no multiple of 8 (size={len(rsp)})'

        statistics = {}
        stat_addr = mem_addr
        for i in range(0, len(rsp), 8):
            read_cnt = struct.unpack('I', rsp[i:i + 4])[0]
            write_cnt = struct.unpack('I', rsp[i + 4:i + 8])[0]
            statistics[stat_addr] = (read_cnt, write_cnt)
            stat_addr += 1
        return statistics

    def info(self) -> Dict[str, Any]:
        log.info('reading device info')
        packet = self.Protocol.packet(self.Protocol.Command.INFO, 0, 0)
        rsp = self._txrx(packet)
        assert len(rsp) > 0, 'received no data'
        info = rsp.decode("utf-8")
        infos = info.split(';')
        assert len(infos) == 4, 'received unexpected number of info segments'
        return dict(
            serial=infos[0],
            firmware=infos[1],
            protocol=infos[2],
            interfaces=int(infos[3]),
        )

    def reset(self) -> None:
        log.info('reset device')
        packet = self.Protocol.packet(self.Protocol.Command.RESET, 0, 0)
        rsp = self._txrx(packet)
        if rsp:
            raise ProtocolError(f'RESET received an unexpected response: {bytestr(rsp)}')



from pathlib import Path
import shlex
import subprocess
from typing import cast, List

from setuptools.command.sdist import sdist as setuptools_sdist
from setuptools.command.build_py import build_py as setuptools_build_py
from setuptools.command.install_lib import install_lib as setuptools_install_lib


class build_py(setuptools_build_py):

    def run(self) -> None:
        for package in self.packages:
            package_files = self.package_data.get(package, [])
            if 'VERSION' not in package_files:
                package_files.append('VERSION')
                self.package_data[package] = package_files
        super().run()


class sdist(setuptools_sdist):

    def make_release_tree(self, base_dir: str, files: List[str]) -> None:
        """Override 'orig.sdist.make_release_tree()' in order to append VERSION files to `sdist``
           distributions.

        The call stack is:
            sdist.run()
                sdist.make_distribution()
                    orig.sdist.make_distribution()
                        orig.sdist.make_release_tree()

        `orig.sdist.make_release_tree()` creates the directory structure for the distribution
        and copies all collected files (from `self.filelist`) into it. The archive is created
        by `orig.sdist.make_distribution()`. Between those, we can add our VERSION files to the
        tree.
        """
        super().make_release_tree(base_dir, files)

        buildpy: setuptools_build_py = cast(
            setuptools_build_py, self.get_finalized_command('build_py'))
        for package in buildpy.packages:
            version_file = Path(base_dir) / package / 'VERSION'
            print(f'writing {version_file}')
            with version_file.open('w+') as file:
                file.write(_describe())


class install_lib(setuptools_install_lib):

    def copy_tree(self,
                  infile: str,
                  outfile: str,
                  preserve_mode: int = 1,
                  preserve_times: int = 1,
                  preserve_symlinks: int = 0,
                  level: int = 1) -> List[str]:
        """Override `install_lib.copy_tree()` in order to generate a VERSION file for each package
           within `infile` directory, for the `bdist_wheel` command. This will then be copied to
           the `outfile` directory by the super method."""
        buildpy: setuptools_build_py = cast(
            setuptools_build_py, self.get_finalized_command('build_py'))
        print(type(outfile))
        for package in buildpy.packages:
            package_path = Path(infile) / package
            if package_path.is_dir():
                version_file = package_path / 'VERSION'
                print(f'creating {version_file}')
                with version_file.open('w+') as file:
                    file.write(_describe())
        return cast(List[str], super().copy_tree(
            infile, outfile, preserve_mode, preserve_times, preserve_symlinks, level))


def _describe(dirty_flag: str = '+dirty') -> str:
    """Attempt to describe the version from `git describe`, or return "unknown."""

    cmd = shlex.split(f'git describe --tags --always --dirty="{dirty_flag}"')
    try:
        proc = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if proc.returncode == 0:
            return proc.stdout.strip().decode()
    except (FileNotFoundError, subprocess.SubprocessError):
        pass
    return 'unknown'


def get(local_flag: str = '', dirty_flag: str = '+dirty') -> str:
    """Finds a version from either the VERSION file (generated by setup.py uppon 'sdist'), or
       from `git describe`.

    The `local_flag` can be passed to mark a version as running locally, when no VERSION file is
    present and `git describe` is used.

    This setup allows for building the correct version in the following cases:
    1. in setup.py on "sdist", `git describe` is used (without a local flag)
    2. in setup.py on "install", the VERSION file is used (previously packaged by "sdist")
    3. when running from a pip installation, the VERSION file is used
    4. when running locally, `git describe` plus a local flag is used, given the following
       definition in the packages __init__.py:
       ```
       __version__ = version.get(local_flag='.local')
       ```
    """
    # FIXME: __package__ will very likely give the wrong package, when version is an own package
    version_file = Path(str(__import__(__package__).__file__)).parent / 'VERSION'
    try:
        with version_file.open('r') as file:
            return file.read().strip()
    except OSError:
        return _describe(dirty_flag) + local_flag



#!/usr/bin/env python3

from pathlib import Path
import shutil

from setuptools import setup

from picoslave import version


def script(name: str, source_path: str) -> str:
    Path(source_path).exists()
    build = Path('build')
    if not build.exists():
        build.mkdir(parents=True, exist_ok=True)
    shutil.copy(source_path, build / name)
    return './' + str(build / name)


setup(
    name='picoslave',
    url='https://gitlab.com/janoskut/picoslave',
    author='Janos Kutscherauer',
    author_email='janoskut@gmail.com',
    license='MIT',
    license_files=['LICENSE'],
    version=version.get(),
    description='PicoSlave is a dual I2C slave simulator for hardware integration testing',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=['python', 'Raspberry Pi Pico', 'I2C', 'testing', 'mock'],
    platforms=['Linux', 'Raspberry Pi'],
    packages=['picoslave'],
    install_requires=['pyusb>=1.2'],
    python_requires='>=3.7',
    cmdclass={
        'build_py': version.build_py,
        'sdist': version.sdist,
        'install_lib': version.install_lib,
    },
    scripts=[script('picoslave-install', 'util/install.sh')],
    entry_points={
        'console_scripts': [
            'picoslave = picoslave.picocli:main',
        ],
    },
)



#!/usr/bin/env python3

"""
\033[1;33mpicocli\033[0m a CLI tool to control the \033[1;33mPicoSlave\033[0m I2C/USB slave simulator.

PicoSlave USB devices have the vid/pid \033[0;33mC001:012C\033[0m and are automatically detected
by this tool if only one is present. The \033[1;39mscan\033[0m command lists devices showing an
index and the device serial number, by which a specific device can be selected
using \033[1;39m-i\033[0m or \033[1;39m-s\033[0m.

\033[1;33mPicoSlave\033[0m supports 2 I2C slaves to be each configured via the \033[1;39mconfig\033[0m command
command for an I2C slave address to operate on a register space of configurable
size starting at address 0x00. Slave operation can be disabled by configuring a
slave for the I2C address 0x00. Slaves 0 or 1 are selected using the \033[1;39miface\033[0m
argument for all slave specific commands.

The register spaces of the slaves can be written into (\033[1;39mwrite\033[0m) and read from (\033[1;39mread\033[0m).
This can be done when slaves are enabled or disabled. \033[1;39mclear\033[0m resets the entire
configured register space to 0x00 or a given value.
"""  # pylint: disable=line-too-long  # noqa

import argparse
import logging
import sys

import picoslave as ps
from picoslave.picoslave import PicoSlave, UsbPicoSlave
from picoslave.picoslave import bytestr
from picoslave.picoslave import DeviceNotFound, UsbError


def config_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    if args.slave_addr == 0:
        print(f'deactivating I2C-{args.iface}')
    else:
        print(f'configuring I2C-{args.iface} for address {args.slave_addr:02X}h, '
              f'size {args.size} and word width {args.width}')
    picoslave.config(args.iface, args.slave_addr, args.size, args.width)


def read_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'reading from I2C-{args.iface}: {args.size} bytes from address {args.addr:02X}h')
    response_data = picoslave.read(args.iface, args.addr, args.size)
    print('response:')
    print(bytestr(response_data))


def write_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'writing to I2C-{args.iface}: {len(args.data)} '
          f'{"bytes" if len(args.data) > 1 else "byte"} to address {args.addr:02X}h:')
    print(bytestr(args.data))
    picoslave.write(args.iface, args.addr, args.data)


def clear_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'clearing memory of I2C-{args.iface} with value {args.value:02x}h')
    picoslave.clear(args.iface, args.value)


def stat_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print(f'reading statistics from I2C-{args.iface}: {args.size} entries '
          f'from address {args.addr:02X}h')
    statistics = picoslave.statistics(args.iface, args.addr, args.size)
    print('  ADDR  READ  WRITE')
    for addr in statistics:
        if statistics[addr][0] or statistics[addr][1]:
            print(f'   {addr:02X}h  {statistics[addr][0]:4}   {statistics[addr][1]:4}')


def info_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    info = picoslave.info()
    print(info)


def reset_cmd(args: argparse.Namespace, picoslave: PicoSlave) -> None:
    print('resetting target')
    try:
        picoslave.reset()
    except UsbError:
        pass


def main() -> None:

    loglevels = ['debug', 'info', 'warning', 'error', 'critical']
    ifaces = [0, 1]

    def auto_int(x: str) -> int:
        return int(x, 0)

    def mem_addr8(x: str) -> int:
        addr = auto_int(x)
        if addr >= 2**8:
            raise argparse.ArgumentTypeError(f'address must be in range [0..{2**8-1}]')
        return addr

    def mem_addr16(x: str) -> int:
        addr = auto_int(x)
        if addr >= 2**16:
            raise argparse.ArgumentTypeError(f'address must be in range [0..{2**16-1}]')
        return addr

    def byte_str(x: str) -> bytes:
        try:
            return bytes.fromhex(x)
        except Exception:
            raise argparse.ArgumentTypeError(f'not a valid byte string: "{x}"')

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-l', '--loglevel', metavar='L', type=str, default='Warning',
                        choices=loglevels,
                        help=f'log level: L={loglevels}')
    parser.add_argument('-d', '--debug', action='store_true',
                        help="shortcut for '--loglevel debug'")
    device_group = parser.add_mutually_exclusive_group()
    device_group.add_argument('-i', '--index', type=int, metavar='X',
                              help="Device index to select a specific PicoSlave device\n"
                                   "(see cmd='scan')")
    device_group.add_argument('-s', '--serial', type=str, metavar='S',
                              help='Device serial number to select a specific PicoSlave\n'
                                   "device (see cmd='scan')")
    parser.add_argument('-V', '--version', action='store_true', help='show the program version')

    # subparsers
    sub = parser.add_subparsers(title='command', dest='command',
                                description='commands to execute on PicoSlave')
    iface_parser = argparse.ArgumentParser(add_help=False)
    iface_parser.add_argument('iface', type=int, metavar='iface', choices=ifaces,
                              help=f'I2C interface number ({ifaces})')

    # config
    config_parser = sub.add_parser('config', aliases=['c'], parents=[iface_parser],
                                   help='configure an I2C slave')
    config_parser.add_argument('slave_addr', type=auto_int,
                               help='Memory or 7-bit slave address (use -8 for 8-bit addresses)')
    config_parser.add_argument('-8', dest='eightbit', action='store_true',
                               help='Set to interpret I2C addresses as 8-bit (default is 7-bit)')
    config_parser.add_argument('size', type=int, nargs='?', default=256,
                               help='size of the configured I2C memory, at which it will overflow '
                                    '(max 256). Note that the I2C memory size is given in words '
                                    '(of size [width]), not bytes.')
    config_parser.add_argument('width', type=int, nargs='?', default=1, choices=[1, 2, 4],
                               metavar='width',
                               help='width of data words in I2C memory (1, 2 or 4)')
    config_parser.set_defaults(func=config_cmd)

    # read
    read_parser = sub.add_parser('read', aliases=['r'], parents=[iface_parser],
                                 help='read from the slaves I2C memory')
    read_parser.add_argument('addr', type=mem_addr16, help='address to start reading from')
    read_parser.add_argument('size', type=int, nargs='?', default=1,
                             help='number of words to read from [addr] (max 256 * [width])')
    read_parser.set_defaults(func=read_cmd)

    # write
    write_parser = sub.add_parser('write', aliases=['w'], parents=[iface_parser],
                                  help='write data to the slaves I2C memory')
    write_parser.add_argument('addr', type=mem_addr16, help='address to start writing into')
    write_parser.add_argument('data', type=byte_str,
                              help='data to write to the I2C memory at [addr].\n'
                                   'Data format is a little endian hex string without spaces, '
                                   'e.g.: 03ab7fE0 (4 bytes). Note that data length must be a '
                                   'multiple of the configured word [width]. (see "config -h")')
    write_parser.set_defaults(func=write_cmd)

    # clear
    clear_parser = sub.add_parser('clear', aliases=['C'], parents=[iface_parser],
                                  help='clear the I2C memory to 0 or an initial value')
    clear_parser.add_argument('value', type=auto_int, nargs='?', default=0x00,
                              help='initialize with this value')
    clear_parser.set_defaults(func=clear_cmd)

    # stat
    stat_parser = sub.add_parser('stat', aliases=['s'], parents=[iface_parser],
                                 help='retrieve read/write statistics from I2C memory')
    stat_parser.add_argument('addr', type=mem_addr8, nargs='?', default=0x00,
                             help='start memory address for reading statistics')
    stat_parser.add_argument('size', type=int, nargs='?', default=256,
                             help='number of addresses/words to get r/w statistics for (max 256)')
    stat_parser.set_defaults(func=stat_cmd)

    # scan
    scan_parser = sub.add_parser('scan', aliases=['S'], help='scan for PicoSlave devices on USB')
    scan_parser.set_defaults(func='scan_cmd')

    # info
    info_parser = sub.add_parser('info', aliases=['I'], help='print info for the selected device')
    info_parser.set_defaults(func=info_cmd)

    # reset
    reset_parser = sub.add_parser('reset', aliases=['R'], help='reset the PicoSlave device')
    reset_parser.set_defaults(func=reset_cmd)

    args = parser.parse_args()

    if args.version:
        print(f'PicoSlave {ps.__version__}')
        exit(0)

    if args.debug:
        args.loglevel = 'debug'

    log_format = '%(asctime)s %(name)-10s %(levelname)-8s %(message)s'
    logging.basicConfig(level=logging.getLevelName(args.loglevel.upper()), format=log_format)

    try:
        import coloredlogs  # type: ignore
        coloredlogs.install(level=logging.getLevelName(args.loglevel.upper()), fmt=log_format)
    except ModuleNotFoundError:
        pass

    if 'func' not in args:
        parser.print_help()
        exit(1)

    if args.func == 'scan_cmd':
        devices = UsbPicoSlave.scan()
        sys.exit(1 if not devices else 0)

    # do command specific parsing before we create the PicoSlave
    if args.func == config_cmd:
        if args.eightbit:
            print(f"converting 8-bit address '{args.slave_addr:02X}h' to 7-bit address "
                  f'{args.slave_addr >> 1:02X}h')
            args.slave_addr = args.slave_addr >> 1
        assert args.slave_addr <= 0x7F, 'I2C address out of range [0..7Fh]'

    assert args.index is None or args.index > 0, 'Device index must not be 0'
    args.index = args.index - 1 if args.index is not None else None

    picoslave = PicoSlave(index=args.index, serial=args.serial)

    try:
        args.func(args, picoslave)
    except (AssertionError, DeviceNotFound) as e:
        if str(e):
            print(f'ERROR: {e}')
            sys.exit(1)
        raise


if __name__ == '__main__':
    main()
